body,html{width:100%;height:100%;margin:0;padding:0;overflow:hidden}body{font-family:-apple-system,BlinkMacSystemFont,Segoe UI,Roboto,Oxygen,Ubuntu,Cantarell,Fira Sans,Droid Sans,Helvetica Neue,sans-serif;-webkit-font-smoothing:antialiased;-moz-osx-font-smoothing:grayscale}code{font-family:source-code-pro,Menlo,Monaco,Consolas,Courier New,monospace}#root{width:100%;height:100%;overflow:hidden}.stats-panel{position:absolute!important}.gfm-color-chip{margin-left:.125rem;display:inline-block;height:.625rem;width:.625rem;border-radius:9999px;border:1px solid gray;transform:TranslateY(.07em)}
